# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Identifies the known communication types and maps them to the appropriate arguments.

Used as a wrapper around switchboard for common communication types and used for detection.
"""
from __future__ import absolute_import
import abc
import os
from gazoo_device import config
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.switchboard import adb_transport
from gazoo_device.switchboard import data_framer
from gazoo_device.switchboard import jlink_transport
from gazoo_device.switchboard import line_identifier
from gazoo_device.switchboard import pty_transport
from gazoo_device.switchboard import serial_transport
from gazoo_device.switchboard import ssh_transport
from gazoo_device.utility import adb_utils
from gazoo_device.utility import host_utils
from gazoo_device.utility import usb_utils
from gazoo_device.utility import usb_config

logger = gdm_logger.get_gdm_logger()

JLINK_COMMS_PRODUCT_NAME = "J-Link"


def get_specific_serial_addresses(match_criteria):
    """Returns addresses for connections that match criteria.

    Args:
      match_criteria (dict): dictionary of match criteria.

    Returns:
      list: list of UsbInfo instances that match criteria.

    Note:
      Excludes Android devices.
      Connections first excluded and then included.
      If there is no include regex, all instances that dont match exclude_regex are included.

      Dictionary entry: {'<key>': 'include_regex': <regex>,
      'exclude_regex': '<regex>'}
    """
    instances = usb_utils.find_matching_connections(match_criteria)
    accessible_instances = [instance for instance in instances
                            if os.access(instance.address, os.R_OK | os.W_OK)]
    inaccessible_addresses = [instance.address for instance in instances
                              if instance not in accessible_instances]
    if inaccessible_addresses:
        logger.info(
            "Warning: no read/write permission for these serial address(es): {}".format(
                inaccessible_addresses))
    return [instance.address for instance in accessible_instances
            if instance.product_name not in usb_config.ANDROID_NAMES]


class CommunicationType():
    """Default settings for a communication type.

    Child classes should at minimum overwrite get_transport_list and get_comms_addresses.
    Child classes will likely overwrite init with their required args.
    """

    def __init__(self, comms_address, **kwargs):
        """Initiates args."""
        self.comms_address = comms_address
        for name, value in kwargs.items():
            setattr(self, name, value)

    def get_button_list(self):
        """Set ups the list of FTDI button instances as needed.

        Notes:
          For non-ftdi devices it just returns an empty list.

        Returns:
          list: list of FTDI button instances.
        """
        return []

    @classmethod
    @abc.abstractmethod
    def get_comms_addresses(cls):
        """Detects the comms addresses that match this communication type.

        Note:
          list can overlap those from other communication types.

        Returns:
          list: list of potential communication addresses.
        """

    def get_data_framers(self, num_transports):
        """Set up framers used to atomicize the raw output of the device.

        Args:
          num_transports (int): number of declared transports.

        Note:
          Deals with interwoven lines as well as tokenized output.
          Default is to use new line character to separate the lines.

        Returns:
           list: list of data framers mapped to each transport.
        """
        return [data_framer.NewlineFramer()] * num_transports

    def get_identifier(self):
        """Setup identifiers used to distinguish loglines from responses.

        Returns:
           LineIdentifier: type of line identifier.

        Notes:
          Defaults to labelling all as unknown.
        """
        return line_identifier.AllUnknownIdentifier()

    def get_partial_line_timeout_list(self, num_transports):
        """Time to wait before returning partial lines for a device.

        Args:
           num_transports (int): number of declared transports.

        Returns:
           list: list of timeouts for each transport.
        """
        return [.1] * num_transports

    def get_switchboard_kwargs(self):
        """Returns the arguments needed to set up switchboard for this communication type.

        Returns:
           dict: transport_list, identifier, framer_list, button_list
        """
        transport_list = self.get_transport_list()
        num_transports = len(transport_list)
        return {
            "transport_list": transport_list,
            "identifier": self.get_identifier(),
            "framer_list": self.get_data_framers(num_transports),
            "button_list": self.get_button_list(),
            "partial_line_timeout_list": self.get_partial_line_timeout_list(num_transports),
        }

    @abc.abstractmethod
    def get_transport_list(self):
        """Setups up the transports which will communicate raw device io.

        Returns:
           list: list of initiated transports.
        """


class AdbComms(CommunicationType):
    """Communication type for devices communicating over ADB.

    Commands/responses go over port 0 while logs are streamed over port 1.
    """

    def __init__(self,
                 comms_address,
                 log_cmd="logcat -v threadtime",
                 shell_cmd="shell",
                 event_log_cmd=None):
        super(AdbComms, self).__init__(comms_address=comms_address,
                                       log_cmd=log_cmd,
                                       shell_cmd=shell_cmd)
        self.event_log_cmd = event_log_cmd

    @classmethod
    def get_comms_addresses(cls):
        return adb_utils.get_adb_devices()

    def get_transport_list(self):
        transports = [
            adb_transport.AdbTransport(comms_address=self.comms_address, command=self.shell_cmd),
            adb_transport.AdbTransport(comms_address=self.comms_address, command=self.log_cmd),
        ]
        if self.event_log_cmd:
            transports.append(adb_transport.AdbTransport(comms_address=self.comms_address,
                                                         command=self.event_log_cmd))
        return transports

    def get_identifier(self):
        return line_identifier.PortLogIdentifier(log_ports=[1, 2])


class DockerComms(CommunicationType):
    """Communication type for VDL Docker devices."""

    def __init__(self, comms_address, secondary_comms_address=None, secondary_log_pattern=None):
        self.comms_address = comms_address
        super().__init__(comms_address=comms_address,
                         secondary_comms_address=secondary_comms_address,
                         secondary_log_pattern=secondary_log_pattern)

    @classmethod
    def get_comms_addresses(cls):
        return host_utils.get_all_vdl_docker_connections()

    def get_transport_list(self):
        docker_exec_args = "exec -i {} /bin/bash".format(self.comms_address)
        docker_logs_args = "logs -f {}".format(self.comms_address)
        return [pty_transport.PtyTransport("docker", docker_exec_args),
                pty_transport.PtyTransport("docker", docker_logs_args)]


class JlinkSerialComms(CommunicationType):
    """Communication type for JLINK and UART communication simultaneously.

    Serial is used for primary communication while jlink is used for log streaming.
    """

    @classmethod
    def get_comms_addresses(cls):
        match_criteria = {"product_name": {"include_regex": JLINK_COMMS_PRODUCT_NAME}}
        return get_specific_serial_addresses(match_criteria)

    def __init__(self, comms_address, baudrate=115200):
        secondary_address = usb_utils.get_serial_number_from_path(comms_address)
        super(JlinkSerialComms, self).__init__(comms_address=comms_address,
                                               secondary_address=secondary_address,
                                               baudrate=baudrate)

    def get_transport_list(self):
        return [serial_transport.SerialTransport(comms_address=self.comms_address,
                                                 baudrate=self.baudrate),
                jlink_transport.JLinkTransport(comms_address=self.secondary_address)]

    def get_identifier(self):
        return line_identifier.PortLogIdentifier()


class PtyProcessComms(CommunicationType):
    """Communication type for communication with a process on the host machine.

    No detection provided.
    """

    @classmethod
    def get_comms_addresses(cls):
        return []

    def __init__(self, comms_address):
        full_command_list = comms_address.split()
        comms_address = full_command_list[0]
        args = " ".join(full_command_list[1:])
        super(PtyProcessComms, self).__init__(comms_address=comms_address, args=args)

    def get_transport_list(self):
        return [pty_transport.PtyTransport(self.comms_address, args=self.args)]

    def get_identifier(self):
        return line_identifier.AllUnknownIdentifier()


class SshComms(CommunicationType):
    """Communication type for device communication over ssh.

    Login is handled by ssh-keys on machine (or device does not require password).

    Commands/responses are on port 0 while log streaming happens on port 1.
    """

    @classmethod
    def get_comms_addresses(cls, static_ips):
        return host_utils.get_all_ssh_ips(static_ips)

    def __init__(self,
                 comms_address,
                 log_cmd="tail -F -n /var/log/messages",
                 args=host_utils.DEFAULT_SSH_OPTIONS,
                 ssh_key_type=None,
                 username="root"):
        super(SshComms, self).__init__(comms_address=comms_address,
                                       log_cmd=log_cmd,
                                       args=args,
                                       username=username,
                                       ssh_key_type=ssh_key_type)

    def get_transport_list(self):
        return [
            ssh_transport.SSHTransport(comms_address=self.comms_address,
                                       args=self.args,
                                       ssh_key_type=self.ssh_key_type,
                                       username=self.username),
            ssh_transport.SSHTransport(comms_address=self.comms_address,
                                       args=self.args,
                                       username=self.username,
                                       ssh_key_type=self.ssh_key_type,
                                       log_cmd=self.log_cmd)]

    def get_identifier(self):
        return line_identifier.PortLogIdentifier(log_ports=[1])


class SerialComms(CommunicationType):
    """Communication type for device communication over basic serial."""

    def __init__(self, comms_address,
                 log_line_regex=None,
                 baudrate=serial_transport.DEFAULT_BAUDRATE):

        super(SerialComms, self).__init__(comms_address=comms_address,
                                          log_line_regex=log_line_regex,
                                          baudrate=baudrate)

    @classmethod
    def get_comms_addresses(cls):
        exclude_products = [JLINK_COMMS_PRODUCT_NAME]
        match_criteria = {
            "product_name": {"exclude_regex": "|".join(exclude_products)},
            "ftdi_interface": {"include_regex": "0"}
        }
        return get_specific_serial_addresses(match_criteria)

    def get_transport_list(self):
        return [serial_transport.SerialTransport(comms_address=self.comms_address,
                                                 baudrate=self.baudrate)]

    def get_identifier(self):
        if self.log_line_regex:
            return line_identifier.RegexLogIdentifier(self.log_line_regex)
        return super(SerialComms, self).get_identifier()


class YepkitComms(CommunicationType):
    """Communication type for yepkit devices.

    Currently supports just detection as communication does not yet work as expected.
    """

    @classmethod
    def get_comms_addresses(cls):
        return host_utils.get_all_yepkit_serials()

    def get_transport_list(self):
        return []


SUPPORTED_CLASSES = [
    AdbComms,
    DockerComms,
    JlinkSerialComms,
    PtyProcessComms,
    SerialComms,
    SshComms,
    YepkitComms]


DETECTION_METHODS_MAP = {a_class.__name__: a_class.get_comms_addresses
                         for a_class in SUPPORTED_CLASSES}
SUPPORTED_TYPES_MAP = {a_class.__name__: a_class for a_class in SUPPORTED_CLASSES}
SUPPORTED_TYPES = list(SUPPORTED_TYPES_MAP.keys())


def detect_connections(static_ips):
    """Detects all the communication addresses for the different devices."""
    connections_dict = {}
    for comms_name, detection_method in DETECTION_METHODS_MAP.items():

        logger.info("\tdetecting potential {} communication addresses".format(comms_name))
        try:
            try:
                comms_addresses = detection_method(static_ips=static_ips)
            except TypeError:  # method does not static static_ips
                comms_addresses = detection_method()
            connections_dict[comms_name] = comms_addresses
            logger.debug(
                "Found {} potential communication addresses".format(
                    len(comms_addresses)))

            # Verify ssh keys exist if ssh connections are detected
            if comms_name == "SshComms" and comms_addresses:
                warnings = []
                missing_keys = []
                ssh_key_types = [key_type for key_type in config.KEYS
                                 if config.KEYS[key_type]["flavor"] == config.KEY_FLAVOR_SSH]
                for ssh_key_type in ssh_key_types:
                    try:
                        host_utils.verify_ssh_key(ssh_key_type)
                    except ValueError as err:
                        warnings.append(str(err))
                    except errors.DownloadKeyError as err:
                        missing_keys.append(err.key_type)
                if warnings:
                    logger.warning("\t\tWARNING: " + "\n\t\t".join(warnings))
                if missing_keys:
                    logger.info(f"WARNING: Found {len(missing_keys)} missing ssh keys: "
                                f"{missing_keys}\nPlease run 'gdm download-keys'. "
                                "See GDM installation documentation for more details")
        except Exception as err:
            logger.info(
                "Unable to detect {} communication addresses. Err: {!r}".format(
                    comms_name, err))
    return connections_dict



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Communication Power Cambrionix Capability."""
import time
from gazoo_device import decorators
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import comm_power_base

logger = gdm_logger.get_gdm_logger()


class CommPowerCambrionix(comm_power_base.CommPowerBase):
    """Base class for comm_power_cambrionix capability."""

    def __init__(self,
                 device_name,
                 manager,
                 data_usb_hub_name,
                 data_usb_port,
                 close_all_transports_fn,
                 open_all_transports_fn,
                 add_log_note_fn,
                 power_and_data_share_cable=False):
        """Create an instance of the comm_power_cambrionix capability.

        Args:
            device_name (str): name of the device this capability is attached to.
            manager (Manager): the manager used to create a device.
            data_usb_hub_name (str): name of the Cambrionix the device data line is connected to.
            data_usb_port (int): Cambrionix port number the device data line is connected to.
            close_all_transports_fn (func): switchboard method to close all transports.
            open_all_transports_fn (func): switchboard method to open all transports.
            add_log_note_fn (func): The method used to send a log note to the device log.
            power_and_data_share_cable (bool): set power_and_data_share_cable to TRUE if device
                                               shares a common USB cable for its power and data.
        Raises:
            ValueError: If the capability cannot be created.
        """
        super().__init__(device_name=device_name)
        if not data_usb_hub_name or not data_usb_port:
            msg = (f"{device_name} the data_usb_hub_name and data_usb_port are unset. "
                   "If the device is plugged into a Cambrionix, issue "
                   "'gdm redetect {device_name}' to set them.")
            logger.debug(msg)
            raise ValueError(msg)
        if not manager:
            msg = (f"{device_name} failed to create comm_power_cambrionix capability."
                   " Device class must not be closed when initiating the capability.")
            logger.debug(msg)
            raise ValueError(msg)

        self._data_usb_hub_name = data_usb_hub_name
        self._data_usb_port = int(data_usb_port)
        self._cambrionix = manager.create_device(data_usb_hub_name)
        self._power_and_data_share_cable = power_and_data_share_cable
        self._add_log_note_fn = add_log_note_fn
        self._close_all_transports_fn = close_all_transports_fn
        self._open_all_transports_fn = open_all_transports_fn

    @decorators.CapabilityLogDecorator(logger)
    def off(self):
        """Turn off power to the device communications port."""
        if self._power_and_data_share_cable:
            if self._add_log_note_fn:
                self._add_log_note_fn(
                    f"comm_power.off() called on {self._device_name} set communication port "
                    f"{self._data_usb_port} to charge as device has a single USB cable "
                    "for data and power.")
            self._close_all_transports_fn()
            self._cambrionix.switch_power.power_on(self._data_usb_port, data_sync=False)
        else:
            self._close_all_transports_fn()
            self._cambrionix.switch_power.power_off(self._data_usb_port)

    @decorators.CapabilityLogDecorator(logger)
    def on(self):
        """Turn on power to the device communication port."""
        self._cambrionix.switch_power.power_on(self._data_usb_port)
        time.sleep(5)  # Small delay to give time for 'dev/tty' to populate
        self._open_all_transports_fn()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for patching Python fire library."""
from __future__ import print_function
import sys
from fire import core
from fire import completion
from fire import inspectutils


def display(lines, out):
    """Display text to user.

    Args:
        lines (str): lines of text to display
        out (object): file object used by interpreter for output

    Note:
        original behavior: lines displayed in scrollable window, similar to "less" shell command.
        behavior with patch: lines printed to the console in standard format.
    """
    if not (out is sys.stdout or out is sys.stderr):
        out = sys.stdout
    text = '\n'.join(lines) + '\n'
    out.write(text)


def visible_members(component, class_attrs=None, verbose=False):
    """Returns a list of the members of the given component.

    Args:
        component (object): the component whose members to list.
        class_attrs (dict): determines how class members will be treated for visibility,
                            defaulting to GetClassAttrsDict.
        verbose (bool): whether to include private members.

    Returns:
      list: of tuples (member_name, member) of all members of the component.

    Note:
        original behavior: component members are collected using inspect.getmembers.
        behavior with patch: component members are collected using _safely_get_members.
    """
    if isinstance(component, dict):
        members = component.items()
    else:
        members = _safely_get_members(component)

    # if class_attrs has not been provided, compute it.
    if class_attrs is None:
        try:
            class_attrs = inspectutils.GetClassAttrsDict(component)  # g3 fire
        except AttributeError:
            class_attrs = completion.GetClassAttrsDict(component)

    return [
        (member_name, member) for member_name, member in members
        if completion.MemberVisible(component, member_name, member, class_attrs=class_attrs,
                                    verbose=verbose)
    ]


def get_member(component, args):
    """Returns a member (function, property, etc.) of a component by consuming CLI arguments.

    Args:
        component (object): the component from which to get the member.
        args (list): arguments from which to consume in the search for the member.

    Returns:
        tuple: (member, consumed_args, remaining_args)
            member: the member that was found by consuming args.
            consumed_args: the args that were consumed by getting this member.
            remaining_args: the remaining args that haven't been consumed yet.

    Raises:
        FireError: if the provided arguments do not correspond to a component member.

    Note:
        original behavior: component members are collected using inspect.getmembers.
        behavior with patch: component members are collected using _safely_get_members.
    """
    arg = args[0]
    arg_names = [
        arg,
        arg.replace('-', '_'),  # treat '-' as '_'.
    ]

    for arg_name in arg_names:
        members = dict(_safely_get_members(component, member_names=[arg_name]))
        if arg_name in members:
            return members[arg_name], [arg], args[1:]

    raise core.FireError('Could not consume arg:', arg)


def apply_patch():
    """Applies monkey patch by attaching custom functions to the Python Fire module."""
    core.Display = display
    core._GetMember = get_member
    completion.VisibleMembers = visible_members


def _safely_get_members(component, member_names=None):
    """Gather stable members of a component.

    Args:
        component (object): The component from which to get the members from.
        member_names (list): List of names of members to gather.

    Returns:
        list: of stable component members as tuples (member name, member instance).

    Note:
        This function was designed to silence exceptions thrown by unstable devices when all
        device properties were previously accessed by inspect.getmembers.
    """
    all_member_names = dir(component)
    member_names = member_names or all_member_names

    members = []
    for attr_name in member_names:
        if attr_name in all_member_names:
            try:
                attribute = getattr(component, attr_name)
                members.append((attr_name, attribute))
            except Exception:
                pass  # ignore members that raise exceptions

    return members



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ADB transport which communicates to the device over a process running 'adb shell'."""
from __future__ import absolute_import
import time

from gazoo_device.switchboard import process_transport
from gazoo_device.utility import adb_utils


READY_RETRY_DELAY = 2


class AdbTransport(process_transport.ProcessTransport):
    """Perform transport communication using the adb executable to some device.
    """

    def __init__(self,
                 comms_address,
                 command="shell",
                 adb_path=None,
                 fastboot_path=None,
                 auto_reopen=True,
                 open_on_start=True):
        """Initialize the AdbTransport object with the given properties.

        Args:
            comms_address (str): device adb_serial to use when connecting to device
            command (str): additional commands to pass to device
            adb_path (str): optional path to adb executable to use
            fastboot_path (str): optional path to fastboot executable to use
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.
        """
        self.comms_address = comms_address
        if adb_path is None:
            adb_path = adb_utils.get_adb_path()
        if fastboot_path is None:
            fastboot_path = adb_utils.get_fastboot_path()
        args = "-s {} {}".format(comms_address, command)
        super(AdbTransport, self).__init__(command=adb_path,
                                           args=args,
                                           auto_reopen=auto_reopen,
                                           open_on_start=open_on_start)
        self._adb_path = adb_path
        self._fastboot_path = fastboot_path

    def _is_ready_to_open(self):
        result = adb_utils.is_device_online(self.comms_address,
                                            adb_path=self._adb_path,
                                            fastboot_path=self._fastboot_path)
        if not result:
            time.sleep(READY_RETRY_DELAY)
        return result



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gets information on all usb connected devices.


Returns a dictionary mapping a persistent address (adb or serial port) to usb info
for that address.

Note: usb_info_mac primarily relies on system_profile for full information.
"""
import json
import os
import subprocess
from gazoo_device import gdm_logger
from gazoo_device.utility import usb_config

logger = gdm_logger.get_gdm_logger()

PROFILER_USB_DATA_TYPE = 'SPUSBDataType'
MAC_COMMAND = ['system_profiler', PROFILER_USB_DATA_TYPE, '-json']
FTDI = 'FTDI'
SERIAL_PATH_TEMPLATE = '/dev/tty.usbserial-{serial_num}{ftdi_line}'


def get_address_to_usb_info_dict():
    """Returns a standard address to usb info dict for mac.

    Returns:
        dict: of address to UsbInfo instance.

    Raises:
        RuntimeError: if system_profiler does not produce the correct output.
    """
    with open(os.devnull, 'w') as devnull:
        output = subprocess.check_output(MAC_COMMAND, stderr=devnull)
    output_dict = json.loads(output)
    usb_connections = output_dict[PROFILER_USB_DATA_TYPE]
    usb_info_dict = {}
    _convert_system_profiler_dict(usb_connections, usb_info_dict)
    return usb_info_dict


def _convert_system_profiler_dict(device_dicts, converted_dicts, port_mapping=None):
    """Convert system profiler dict to the system agnostic form.

    Recursively walks the tree of connected USB devices.

    Args:
        device_dicts (list): dictionaries from system_profiler to convert to a generic form.
        converted_dicts (dict): converted dictionaries will be stored here.
        port_mapping (dict): mapping from USB device number (str) to Cambrionix port number (int).
    """
    for device_dict in device_dicts:
        if device_dict.get('manufacturer') == 'cambrionix':
            children = _create_device_entries_for_cambrionix(device_dict, device_dict['_name'])
            converted_dicts.update(children)
        else:
            if 'serial_num' in device_dict:
                entry = _create_device_entry(device_dict, port_mapping)
                converted_dicts.update(entry)
            _convert_system_profiler_dict(device_dict.get("_items", []),
                                          converted_dicts,
                                          port_mapping)


def _add_child_addresses(children_dicts):
    """Add child addresses to cambrionix entry and cambrionix to children."""
    child_addresses = list(children_dicts.keys())
    cambrionix_address = [key for key, value in children_dicts.items()
                          if value.product_name == usb_config.CAMBRIONIX_NAME][0]
    children_dicts[cambrionix_address].child_addresses = child_addresses
    for address in child_addresses:
        if address != cambrionix_address:
            children_dicts[address].usb_hub_address = cambrionix_address


def _create_device_entries_for_cambrionix(device_dict, cambrionix_model):
    """Converts device dict for cambrionix into individual entries for children."""
    port_mapping = usb_config.CAMBRIONIX_PORT_MAP[cambrionix_model]
    children_dicts = {}
    for entry in device_dict['_items']:
        _convert_system_profiler_dict(entry.get('_items', []),
                                      children_dicts,
                                      port_mapping)
    _add_child_addresses(children_dicts)
    return children_dicts


def _create_device_entry(device_dict, port_mapping=None):
    """Creates a standard usb_info entry for a device."""
    return_dict = {}
    address = SERIAL_PATH_TEMPLATE.format(serial_num=device_dict['serial_num'], ftdi_line='')
    usb_entry = usb_config.UsbInfo()
    usb_entry.serial_number = device_dict['serial_num']
    usb_entry.manufacturer = device_dict['manufacturer']
    usb_entry.product_id = _get_product_id(device_dict)
    usb_entry.vendor_id = _get_vendor_id(device_dict)
    # Product names are reported inconsistently between gLinuxes and Macs; either undescores or
    # spaces are used. Convert all underscores to spaces as a workaround.
    usb_entry.product_name = device_dict['_name'].replace("_", " ")
    usb_entry.usb_hub_port = _get_port_number(device_dict, port_mapping)
    if 'Media' in device_dict:  # mounted device

        if 'volumes' in device_dict['Media'][0]:
            volume = device_dict['Media'][0]['volumes'][0]
            address = '/dev/' + volume["bsd_name"]
            usb_entry.disk = address
    if usb_entry.product_name in usb_config.ANDROID_NAMES:
        usb_entry.address = usb_entry.serial_number  # android serial
        return_dict[usb_entry.address] = usb_entry
    else:
        usb_entry.address = address
        if os.path.exists(usb_entry.address):
            return_dict[usb_entry.address] = usb_entry
    return return_dict


def _get_product_id(device_dict):
    """Converts 0x1234 to 1234."""
    return device_dict['product_id'].split('x')[-1]


def _get_vendor_id(device_dict):
    """Converts '0x1234 Serial Interational' to 1234."""
    return device_dict['vendor_id'].split()[0].split('x')[-1]


def _get_port_number(device_dict, port_mapping=None):
    """Last two digits correspond to the port number.

    Args:
      device_dict (dict): device info from system profiler.
      port_mapping (dict): map of port identifiers to port numbers.

    Returns:
       int: port number or None if no cambrionix port applicable..

    Examples:
        "0x14443000 / 10" -> port identifier is "43" -> port 5 on PP15S.
        "0x14313400" -> port identifier is "34" -> port 8 on PP15S.
    """
    if not port_mapping:  # not connected to a usb hub
        return None
    if is_cambrionix(device_dict):
        return None
    # get the appropriate digits
    stripped_location_id = device_dict['location_id'].split()[0].rstrip("0")
    index = "{}.{}".format(stripped_location_id[-2], stripped_location_id[-1])
    port = port_mapping[index]
    return port


def is_cambrionix(device_dict):
    """Device dict is a cambrionix entry."""
    return device_dict['_name'] == usb_config.CAMBRIONIX_NAME



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for local host commands."""
from __future__ import absolute_import
import os
import re
import shutil
import subprocess

from gazoo_device import config
from gazoo_device import errors
from gazoo_device import gdm_logger

logger = gdm_logger.get_gdm_logger()
ARP_CONNECTED_IPS = r'([\-\w\.]*)\s*ether'
DOCKER_CP_COMMAND = "cp {src} {dest}"
DOWNLOAD_KEY_TIMEOUT = 3
PACKAGE_KEY_DIR = os.path.join(config.PACKAGE_PATH, "keys")
GSUTIL_CLI = None
IP_ADDRESS = r"\d+\.\d+\.\d+\.\d+"
PING_COMMAND = "ping -c 1 -W 2 {}"  # Try to ping ip address and wait for reponse for 2 seconds.
SSHABLE_COMMAND = "nc -z -w 2 {} 22"  # Try to connect to ssh port for up to 2 seconds.

SSH_ARGS = "{options} {user}@{ip_address} {command}"
SCP_COMMAND = "scp {ssh_opt} {src} {dest}"
SSH_TIMEOUT = 3
SSH_CONFIG = ("-oPasswordAuthentication=no -oStrictHostKeyChecking=no -oBatchMode=yes "
              "-oConnectTimeout={timeout}".format(timeout=SSH_TIMEOUT))
_SSH_DISABLE_PSEUDO_TTY = "-T "
DEFAULT_SSH_OPTIONS = _SSH_DISABLE_PSEUDO_TTY + SSH_CONFIG

GET_COMMAND_PATH = "which {}"
GET_CONNECTED_IPS = "/usr/sbin/arp -e"

GSUTIL_DICT = {
}


class _SCP_DIRECTION(object):
    TO_DEVICE = "to_device"
    FROM_DEVICE = "from_device"


def docker_cp_to_device(docker_container,
                        local_file_path,
                        container_file_path):
    """Send a file to a docker container using a docker cp command.

    Args:
        docker_container (str): docker container identifier
        local_file_path (str): path to file on host.
        container_file_path (str): path to destination on device.

    Returns:
        str: command output.
    """
    source, destination = local_file_path, "{}:{}".format(docker_container, container_file_path)
    cmd = DOCKER_CP_COMMAND.format(src=source, dest=destination)
    return _execute_docker_cmd(cmd)


def docker_cp_from_device(docker_container,
                          local_file_path,
                          container_file_path):
    """Receive file from a docker container using a docker cp command.

    Args:
        docker_container (str): docker container identifier
        local_file_path (str): path to file on host.
        container_file_path (str): path to destination on device.

    Returns:
        str: command output.
    """
    destination, source = local_file_path, "{}:{}".format(docker_container, container_file_path)
    cmd = DOCKER_CP_COMMAND.format(src=source, dest=destination)
    return _execute_docker_cmd(cmd)


def download_key(key_type):
    """Download an GDM key (ssh, api, etc.) if it doesn't currently exist locally.

    Args:
        key_type (str): type of key to download.

    Raises:
        DownloadKeyError: if unable to download the key.
    """
    local_path = config.KEYS[key_type]["local_path"]
    key_flavor = config.KEYS[key_type]["flavor"]
    if not os.path.exists(local_path):
        try:
            _obtain_key(key_type)
        except Exception as err:
            raise errors.DownloadKeyError(key_type, [err])
        logger.info(f"{key_type} {key_flavor} key downloaded")
    else:
        logger.debug(f"{key_type} {key_flavor} key found")


def generate_ssh_args(ip_address, command, user, options=DEFAULT_SSH_OPTIONS, ssh_key_type=None):
    """Generate list of ssh args to send to subprocess.

    Args:
      ip_address (str): ip address to ssh to.
      command (str): command to send.
      user (str): user for ip address.
      options (str): extra args for ssh command.
      ssh_key_type (str): type of ssh key

    Returns:
        str: str of args to send to subprocess.
    """
    if ssh_key_type:
        verify_ssh_key(ssh_key_type)
        options += " -i " + config.KEYS[ssh_key_type]["local_path"]
    return SSH_ARGS.format(options=options, ip_address=ip_address, user=user,
                           command=command)


def get_command_path(command_name):
    """Return the full path for the given command name.

    Args:
        command_name (str): Name of the command to look for. Example: 'fastboot'

    Raises:
        CalledProcessError: Retrieving the command name path failed.

    Returns:
        str: Output for the given command name.
    """
    try:
        result = subprocess.check_output(
            ['which', command_name], stderr=subprocess.STDOUT).rstrip()
        return result.decode("utf-8", "replace")
    except subprocess.CalledProcessError:
        return ""


def get_all_connected_arp_ips():
    """Gets all connected ips from arp -e list.

    Returns:
        list: list of ip addresses.

    Note:
       Does not return ip_addresses listed as 'incomplete'.
    """
    try:
        output = subprocess.check_output([u"/usr/sbin/arp", u"-e"], stderr=subprocess.STDOUT)
        addresses = re.findall(ARP_CONNECTED_IPS, output.decode("utf-8", "replace"), re.MULTILINE)
        return addresses
    except subprocess.CalledProcessError:
        return []


def is_static_ip(comm_port):
    return re.search(IP_ADDRESS, comm_port)


def get_all_ssh_ips(static_ips=None):
    """Returns all ips that are not ISA3."""
    if not static_ips:
        static_ips = []
    pingable_static_ips = [ip for ip in static_ips if is_pingable(ip)]
    unpingable_static_ips = [ip for ip in static_ips if ip not in pingable_static_ips]
    if unpingable_static_ips:
        logger.info("Warning: ip_address(es) provided are not pingable: {}".format(
            unpingable_static_ips))
    possible_ips = sorted(set(pingable_static_ips))
    sshable_ips = [ip for ip in possible_ips if is_sshable(ip)]
    unsshable_ips = [ip for ip in possible_ips if ip not in sshable_ips]
    if unsshable_ips:
        logger.info(
            "Warning: ip_address(es) provided do not have ssh port 22 open: {}".format(
                unsshable_ips))
    return sshable_ips


def get_all_yepkit_serials():
    """Returns all Yepkit serials."""
    try:
        if not has_command("ykushcmd"):
            logger.debug("Ykushcmd is not installed on machine. Cannot get yepkit serials.")
            return []

        results = subprocess.check_output(["ykushcmd", "-l"], stderr=subprocess.STDOUT)
        results = results.decode("utf-8", "replace")
        logger.debug("get_all_yepkit_serials returned: {!r}".format(results))
    except subprocess.CalledProcessError as err:
        logger.debug("Checking get_all_yepkit_serial returned err: {!r}".format(err))
        return []

    # If no YKUSH boards are found then the YK21624 line will be this instead:
    #    No YKUSH boards found.
    if "No YKUSH boards found" in results:
        return []

    # Typical ykushcmd -l output looks like this:
    #    Attached YKUSH Boards:\n\nYK21624\n\nYK21623

    # Removes blank lines and header
    results = [line for line in results.splitlines() if line]
    return results[1:]


def get_all_vdl_docker_connections():
    """Returns all VDL docker ids."""
    try:
        cmd = ['docker', 'ps', '--filter', 'name=VDL', '--format', '{{.ID}}']
        results = subprocess.check_output(cmd)
    except subprocess.CalledProcessError as err:
        logger.debug("Checking get_all_vdl_docker_connections returned err: {!r}".format(err))
        return []
    return results.decode().splitlines()


def gsutil_command(cmd, gsutil_path, extra_args=None):
    """Send a gsutil command using a valid gsutil path.

    Args:
      cmd (str): gsutil command.
      gsutil_path (str): path in gsutil directory.
      extra_args (list): any other args needed.

    Returns:
      str: output of command.

    Raises:
      RuntimeError: if unable to send command or find valid gsutil.
    """
    _set_gsutil_cli()
    _set_boto_file(gsutil_path)

    cmd_list = [GSUTIL_CLI, cmd, gsutil_path]
    if extra_args:
        cmd_list += extra_args

    try:
        output = subprocess.check_output(cmd_list, stderr=subprocess.STDOUT)
        return output.decode("utf-8", "replace")
    except subprocess.CalledProcessError as err:
        raise RuntimeError("{!r} failed. Err: {}".format(" ".join(cmd_list), err.output))


def has_command(command_name):
    """Determine if the given command executable is present on the local host.

    Args:
        command_name (str): Name of the command to look for. Example: 'fastboot'

    Returns:
        bool: True if the command is found in the user's $PATH, False otherwise.
    """

    result = get_command_path(command_name)
    return bool(result)


def is_in_arp_table(ip_address):
    """Determine if the given IP address is present in arp table.

    Args:
        ip_address (str): to check for in arp table

    Returns:
        bool: True if IP or MAC address is in arp table, False otherwise.

    Note:
        The MAC address provided must be in the format of '##:##:##:##:##:##'
        for this method to work correctly.
    """
    return ip_address in get_all_connected_arp_ips()


def is_in_ifconfig(ip_or_mac_address):
    """Determine if the given IP or MAC address is present in ifconfig output.

    Args:
        ip_or_mac_address (str): to check for in ifconfig output

    Returns:
        bool: True if IP or MAC address is in ifconfig output, False otherwise.

    Note:
        The MAC address provided must be in the format of '##:##:##:##:##:##'
        for this method to work correctly.
    """
    try:
        output = subprocess.check_output(["ifconfig"], stderr=subprocess.STDOUT)
        return ip_or_mac_address in output.decode("utf-8", "replace")
    except subprocess.CalledProcessError:
        return False


def is_pingable(ip_address):
    """Determine if the given IP address is pingable.

    Args:
        ip_address (str): to ping

    Returns:
        bool: True if IP address is pingable with no loss within 1 second.
    """
    try:
        cmd_list = PING_COMMAND.format(ip_address).split()
        subprocess.check_output(cmd_list, stderr=subprocess.STDOUT)
        return True
    except subprocess.CalledProcessError:
        return False


def is_sshable(ip_address):
    """Determine if the given IP address has an open ssh port.

    Args:
        ip_address (str): to ping

    Returns:
        bool: True if nc can see port 22 open.
    """
    try:
        cmd_list = SSHABLE_COMMAND.format(ip_address).split()
        subprocess.check_output(cmd_list, stderr=subprocess.STDOUT)
        return True
    except subprocess.CalledProcessError:
        return False


def ssh_command(ip_address, command, user="root", options=DEFAULT_SSH_OPTIONS, ssh_key_type=None):
    """Send an ssh command.

    Args:
        ip_address (str): device
        command (str): command to send to device.
        user (str): user to login to ip.
        options (str): extra args for ssh command.
        ssh_key_type (str): which ssh key to use. Key types stored in config.

    Returns:
        str: output

    Raises:
       RuntimeError: if ssh command fails.
    """
    ssh_args = generate_ssh_args(ip_address, command, user, options, ssh_key_type)
    ssh_list = ["ssh"] + ssh_args.split()
    try:
        result = subprocess.check_output(ssh_list, stderr=subprocess.STDOUT)
        result = result.decode("utf-8", "replace")
        logger.debug("Ssh command {} to {} returned {!r}".format(command, ip_address, result))
        return result
    except subprocess.CalledProcessError as err:
        msg = "Command {} failed. Err: {!r}".format(" ".join(ssh_list), err.output)
        logger.debug(msg)
        raise RuntimeError(msg)


def _scp(source,
         destination,
         options=SSH_CONFIG,
         ssh_key_type=None):
    """Send file to or from the device using "scp" utility.

    Args:
        source (str): source file.
        destination (str): where to copy te file to.
        options (str): SSH options to pass to scp.
        ssh_key_type (str): type of ssh key

    Returns:
        str: command output.

    Raises:
        ValueError: invalid direction value provided.
        RuntimeError: if scp fails.
    """
    if ssh_key_type:
        verify_ssh_key(ssh_key_type)
        options += " -i " + config.KEYS[ssh_key_type]["local_path"]

    command = SCP_COMMAND.format(ssh_opt=options, src=source, dest=destination)
    try:
        logger.debug("Executing {!r}".format(command))
        result = subprocess.check_output(command.split(), stderr=subprocess.STDOUT)
        return result.decode("utf-8", "replace")
    except subprocess.CalledProcessError as err:
        raise RuntimeError(
            "Scp {!r} failed. Error: {!r}. Output: {}.".format(command, err, err.output))


def scp_to_device(ip_address,
                  local_file_path,
                  remote_file_path,
                  user="root",
                  options=SSH_CONFIG,
                  ssh_key_type=None):
    """Send file to device via scp.

    Args:
        ip_address (str): ip_address
        local_file_path (str): source file.
        remote_file_path (str): where to copy te file to.
        user (str): username for scp to use on the device.
        options (str): SSH options to pass to scp.
        ssh_key_type (str): type of ssh key.

    Returns:
        str: command output.
    """
    remote_file_path = "{user}@{host}:{path}".format(user=user,
                                                     host=ip_address,
                                                     path=remote_file_path)
    return _scp(source=local_file_path,
                destination=remote_file_path,
                options=options,
                ssh_key_type=ssh_key_type)


def scp_from_device(ip_address,
                    local_file_path,
                    remote_file_path,
                    user="root",
                    options=SSH_CONFIG,
                    ssh_key_type=None):
    """Send file from device via scp.

    Args:
        ip_address (str): ip_address
        local_file_path (str): where to copy the file to.
        remote_file_path (str): source file.
        user (str): username for scp to use on the device.
        options (str): SSH options to pass to scp.
        ssh_key_type (str): type of ssh key.

    Returns:
        str: command output.
    """
    remote_file_path = "{user}@{host}:{path}".format(user=user,
                                                     host=ip_address,
                                                     path=remote_file_path)
    return _scp(source=remote_file_path,
                destination=local_file_path,
                options=options,
                ssh_key_type=ssh_key_type)


def verify_ssh_key(ssh_key_type):
    """Verifies ssh key exists and has permissions correctly set.

    Args:
        ssh_key_type (str): device type for error message purposes.

    Raises:
        GazooDeviceError: if ssh key inaccessible.
        ValueError: if unable to set permissions of ssh key.
    """
    local_path = config.KEYS[ssh_key_type]["local_path"]
    if not os.path.isfile(local_path):
        download_key(ssh_key_type)
        if not os.path.isfile(local_path):
            raise errors.GazooDeviceError(
                f"{ssh_key_type} ssh key missing. Please run 'gdm download-keys'")

    permissions = oct(os.stat(local_path).st_mode)[-3:]
    if not (permissions == "400" or permissions == "600"):
        try:
            os.chmod(local_path, 0o400)
        except OSError:
            raise ValueError(
                "Unable to communicate to {} ssh key. Permissions are {}."
                "'chmod 400 {}'".format(
                    ssh_key_type,
                    permissions,
                    local_path))


def _set_gsutil_cli():
    """Finds a valid gsutil executable and sets the module-level GSUTIL_CLI constant.

    Raises:
        RuntimeError: unable to find a valid "gsutil" binary.
    """
    global GSUTIL_CLI
    if not GSUTIL_CLI:
        possible_clis = [
            get_command_path("gsutil"),
            "/usr/local/bin/gsutil"
        ]
        for cli in possible_clis:
            if os.path.exists(cli):
                logger.debug("Setting {} to be the gsutil cli".format(cli))
                GSUTIL_CLI = cli
                return
        raise RuntimeError("Unable to find valid gsutil binary."
                           "Refer to GDM installation documentation.")


def _choose_boto_file(path):
    """Choose the correct .boto file for the given gsutil path.

    Args:
        path (str): gs address of file to copy.

    Raises:
       RuntimeError: if no relevant boto file found.

    Returns:
      str: boto_file_path to use.
    """
    for prefix, values in GSUTIL_DICT.items():
        if prefix in path:
            if values["verified"]:
                return values["boto_file"]
            elif os.path.exists(values["boto_file"]):
                values["verified"] = True
                return values["boto_file"]
            elif os.path.exists(config.DEFAULT_BOTO):
                logger.warning("WARNING: Unable to locate appropriate GDM boto file ({})."
                               " Using default .boto file ({}) for access."
                               .format(values["boto_file"], config.DEFAULT_BOTO))
                values["verified"] = True
                values["boto_file"] = config.DEFAULT_BOTO
                return config.DEFAULT_BOTO

    if os.path.exists(config.DEFAULT_BOTO):
        logger.warning("Unrecognized gsutil location. Using default .boto file ({}) for access."
                       .format(config.DEFAULT_BOTO))
        return config.DEFAULT_BOTO
    raise RuntimeError(
        "No .boto file {} found to give access to gcs storage."
        " Has 'gsutil config' been run to set up gcs access?".format(config.DEFAULT_BOTO))


def _obtain_key(key_type):
    """Downloads an GDM key (ssh, api, etc.) and stores it in a local file.

    Args:
        key_type (str): type of key to download.
    """
    local_path = config.KEYS[key_type]["local_path"]
    remote_filename = config.KEYS[key_type]["remote_filename"]
    shutil.copyfile(src=os.path.join(PACKAGE_KEY_DIR, remote_filename),
                    dst=local_path)
    # The keys included with GDM are not secret, so it's okay to
    # include them in the repo for the GDM prototype.
    # TODO: package keys somewhere else (like GCS?).


def _set_boto_file(path):
    """Sets the appropriate boto file for that gs path.

    Args:
      path (str): gs address of file to copy.

    Raises:
       RuntimeError: if no boto file exists.
    """
    boto_file = _choose_boto_file(path)

    if os.path.exists(boto_file):
        logger.debug("Using {} to access {}".format(boto_file, path))
        os.environ["BOTO_CONFIG"] = boto_file
    else:
        raise RuntimeError("Expecting .boto {boto_file} to exist to access {path} but it doesn't."
                           " Copy ~/.boto to {boto_file} if you have access on the "
                           "gsutil command line".format(boto_file=boto_file, path=path))


def _execute_docker_cmd(cmd):
    """Send file to or from the device using "docker cp" utility.

    Args:
        cmd (str): command

    Returns:
        str: command output.

    Raises:
        RuntimeError: If unable to run the command
    """

    full_cmd = "docker {}".format(cmd)
    try:
        logger.debug("Executing {!r}".format(full_cmd))
        result = subprocess.check_output(full_cmd.split(), stderr=subprocess.STDOUT)
        return result.decode("utf-8", "replace")
    except subprocess.CalledProcessError as err:
        raise RuntimeError(
            "Docker command: {!r} failed, with error: {!r}. Output: {}.".format(full_cmd,
                                                                                err,
                                                                                err.output))
